import collections
from fractions import Fraction
try:
    from quicktions import Fraction # type: ignore
except ImportError:
    pass
import functools
import math
import typing
from abjad import enums
from abjad import exceptions
from abjad import mathtools
from abjad import typings
from abjad.markups import Markup
from abjad.markups import MarkupCommand
from abjad.mathtools.NonreducedFraction import NonreducedFraction
from abjad.mathtools.Ratio import Ratio
from abjad.scheme import Scheme
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.system.StorageFormatManager import StorageFormatManager
from abjad.top.new import new
from abjad.top.sequence import sequence
from abjad.utilities.Duration import Duration
from abjad.utilities.Enumerator import Enumerator
from abjad.utilities.Multiplier import Multiplier
from .LilyPondLiteral import LilyPondLiteral


@functools.total_ordering
class MetronomeMark(AbjadValueObject):
    r"""
    MetronomeMark.

    ..  container:: example

        Initializes integer-valued metronome mark:

        >>> score = abjad.Score([])
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), 90)
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \tempo 4=90
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    ..  container:: example

        Initializes rational-valued metronome mark:

        >>> score = abjad.Score([])
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), abjad.Fraction(181, 2))
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \tempo \markup {
                        \scale
                            #'(0.75 . 0.75)
                            \score
                                {
                                    \new Score
                                    \with
                                    {
                                        \override SpacingSpanner.spacing-increment = #0.5
                                        proportionalNotationDuration = ##f
                                    }
                                    <<
                                        \new RhythmicStaff
                                        \with
                                        {
                                            \remove Time_signature_engraver
                                            \remove Staff_symbol_engraver
                                            \override Stem.direction = #up
                                            \override Stem.length = #5
                                            \override TupletBracket.bracket-visibility = ##t
                                            \override TupletBracket.direction = #up
                                            \override TupletBracket.minimum-length = #4
                                            \override TupletBracket.padding = #1.25
                                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                            \override TupletNumber.font-size = #0
                                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                            tupletFullLength = ##t
                                        }
                                        {
                                            c'4
                                        }
                                    >>
                                    \layout {
                                        indent = #0
                                        ragged-right = ##t
                                    }
                                }
                        =
                        \raise
                            #-0.5
                            {
                                90
                                \tiny
                                    \fraction
                                        1
                                        2
                            }
                        }
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    ..  container:: example

        Initializes from text, duration and range:

        >>> score = abjad.Score([])
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> mark = abjad.MetronomeMark((1, 4), (120, 133), 'Quick')
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \tempo Quick 4=120-133
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

    ..  container:: example

        Use rational-value units-per-minute together with custom markup for
        float-valued metornome marks:

        >>> score = abjad.Score([])
        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> score.append(staff)
        >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
        ...     abjad.Duration(1, 4),
        ...     90.1,
        ...     )
        >>> mark = abjad.MetronomeMark(
        ...     (1, 4),
        ...     Fraction(900, 10),
        ...     custom_markup=markup,
        ...     )
        >>> abjad.attach(mark, staff[0])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \tempo \markup {
                        \scale
                            #'(0.75 . 0.75)
                            \score
                                {
                                    \new Score
                                    \with
                                    {
                                        \override SpacingSpanner.spacing-increment = #0.5
                                        proportionalNotationDuration = ##f
                                    }
                                    <<
                                        \new RhythmicStaff
                                        \with
                                        {
                                            \remove Time_signature_engraver
                                            \remove Staff_symbol_engraver
                                            \override Stem.direction = #up
                                            \override Stem.length = #5
                                            \override TupletBracket.bracket-visibility = ##t
                                            \override TupletBracket.direction = #up
                                            \override TupletBracket.minimum-length = #4
                                            \override TupletBracket.padding = #1.25
                                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                            \override TupletNumber.font-size = #0
                                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                            tupletFullLength = ##t
                                        }
                                        {
                                            c'4
                                        }
                                    >>
                                    \layout {
                                        indent = #0
                                        ragged-right = ##t
                                    }
                                }
                        =
                        \general-align
                            #Y
                            #-0.5
                            90.1
                        }
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>

        (Abjad models all timekeeping with rationals.)

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_custom_markup',
        '_hide',
        '_reference_duration',
        '_textual_indication',
        '_units_per_minute',
        )

    _context = 'Score'

    _format_slot = 'opening'

    _parameter = 'METRONOME_MARK'

    ### INITIALIZER ###

    def __init__(
        self,
        reference_duration: typing.Union[
            Duration, typing.Tuple[int, int]] = None,
        units_per_minute: typings.Number = None,
        textual_indication: Markup = None,
        *,
        custom_markup: Markup = None,
        hide: bool = None,
        ) -> None:
        assert isinstance(textual_indication, (str, type(None)))
        arguments = (reference_duration, units_per_minute, textual_indication)
        if all(_ is None for _ in arguments):
            reference_duration = (1, 4)
            units_per_minute = 60
        if reference_duration:
            reference_duration = Duration(reference_duration)
        if isinstance(units_per_minute, float):
            raise Exception(
                f'do not set units-per-minute to float ({units_per_minute});'
                ' use fraction with textual indication instead.'
                )
        prototype = (
            int,
            Fraction,
            collections.Sequence,
            type(None),
            )
        assert isinstance(units_per_minute, prototype)
        if isinstance(units_per_minute, collections.Sequence):
            assert len(units_per_minute) == 2
            item_prototype = (int, Duration)
            assert units_per_minute is not None
            assert all(isinstance(x, item_prototype) for x in units_per_minute)
            units_per_minute = tuple(sorted(units_per_minute))
        self._reference_duration = reference_duration
        self._textual_indication = textual_indication
        self._units_per_minute = units_per_minute
        if custom_markup is not None:
            assert isinstance(custom_markup, Markup), repr(custom_markup)
        self._custom_markup = custom_markup
        if hide is not None:
            hide = bool(hide)
        self._hide = hide

    ### SPECIAL METHODS ###

    def __add__(self, argument) -> typing.Optional['MetronomeMark']:
        """
        Adds metronome mark to ``argument``.

        ..  container:: example

            Adds one metronome mark to another:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

            >>> mark_2 + mark_1
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

        ..  container:: example

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example

            Raises imprecise metronome mark error with range:

            >>> mark_1 = abjad.MetronomeMark((1, 8), (90, 92))
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example

            Raises type error when ``argument`` is not a metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60) + 90
            Traceback (most recent call last):
                ...
            TypeError: 90

        """
        if not isinstance(argument, type(self)):
            raise TypeError(argument)
        if self.is_imprecise or argument.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, Fraction)
        assert isinstance(argument.quarters_per_minute, Fraction)
        assert isinstance(self.reference_duration, Duration)
        assert isinstance(argument.reference_duration, Duration)
        new_quarters_per_minute = \
            self.quarters_per_minute + argument.quarters_per_minute
        minimum_denominator = min((
            self.reference_duration.denominator,
            argument.reference_duration.denominator,
            ))
        nonreduced_fraction = NonreducedFraction(new_quarters_per_minute / 4)
        nonreduced_fraction = nonreduced_fraction.with_denominator(
            minimum_denominator)
        new_units_per_minute, new_reference_duration_denominator = \
            nonreduced_fraction.pair
        new_reference_duration = Duration(
            1,
            new_reference_duration_denominator
            )
        metronome_mark = type(self)(
            new_reference_duration,
            new_units_per_minute,
            )
        return metronome_mark

    def __div__(self, argument) -> 'MetronomeMark':
        """
        Divides metronome mark by ``argument``.

        ..  container:: example

            Divides metronome mark by number:

            >>> abjad.MetronomeMark((1, 4), 60) / 2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Divides metronome mark by other metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60) / abjad.MetronomeMark((1, 4), 40)
            Multiplier(3, 2)

        """
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        if getattr(argument, 'is_imprecise', False):
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, Fraction)
        if isinstance(argument, type(self)):
            assert isinstance(argument.quarters_per_minute, Fraction)
            result = self.quarters_per_minute / argument.quarters_per_minute
            return Multiplier(result)
        elif isinstance(argument, (int, Fraction)):
            assert isinstance(self.units_per_minute, (int, Fraction))
            units_per_minute = self.units_per_minute / argument
            if mathtools.is_integer_equivalent_number(units_per_minute):
                units_per_minute = int(units_per_minute)
            else:
                units_per_minute = Fraction(units_per_minute)
            result = new(self, units_per_minute=units_per_minute)
            return result
        else:
            raise TypeError(f'must be number or metronome mark: {argument!r}.')

    def __eq__(self, argument) -> bool:
        """
        Is true when metronome mark equals ``argument``.

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52)
            >>> mark_2 = abjad.MetronomeMark((3, 32), 52)

            >>> mark_1 == mark_2
            True

            >>> mark_2 == mark_1
            True

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52)
            >>> mark_2 = abjad.MetronomeMark((6, 32), 104)

            >>> mark_1 == mark_2
            False

            >>> mark_2 == mark_1
            False

        ..  container:: example

            >>> mark_1 = abjad.MetronomeMark((3, 32), 52, 'Langsam')
            >>> mark_2 = abjad.MetronomeMark((3, 32), 52, 'Langsam')
            >>> mark_3 = abjad.MetronomeMark((3, 32), 52, 'Slow')

            >>> mark_1 == mark_2
            True

            >>> mark_1 == mark_3
            False

        """
        if not isinstance(argument, type(self)):
            return False
        # 'hide' is excluded from equality testing:
        if (self.reference_duration == argument.reference_duration and
            self.units_per_minute == argument.units_per_minute and
            self.textual_indication == argument.textual_indication and
            self.custom_markup == argument.custom_markup):
            return True
        return False

    def __format__(self, format_specification='') -> str:
        r"""
        Formats metronome mark.

        ..  container:: example

            Without custom markup:

            >>> mark = abjad.MetronomeMark((1, 4), 84, 'Allegro')
            >>> print(format(mark))
            abjad.MetronomeMark(
                reference_duration=abjad.Duration(1, 4),
                units_per_minute=84,
                textual_indication='Allegro',
                )

        ..  container:: example

            With custom markup:

            >>> markup = abjad.Markup(r'\italic { Allegro }')
            >>> mark = abjad.MetronomeMark((1, 4), 84, custom_markup=markup)
            >>> print(format(mark))
            abjad.MetronomeMark(
                reference_duration=abjad.Duration(1, 4),
                units_per_minute=84,
                custom_markup=abjad.Markup(
                    contents=[
                        abjad.MarkupCommand(
                            'italic',
                            ['Allegro']
                            ),
                        ],
                    ),
                )

        """
        if format_specification in ('', 'storage'):
            return StorageFormatManager(self).get_storage_format()
        assert format_specification == 'lilypond'
        return self._get_lilypond_format()

    def __hash__(self) -> int:
        """
        Hashes metronome mark.

        Redefined in tandem with __eq__.
        """
        return AbjadValueObject.__hash__(self)

    def __lt__(self, argument) -> bool:
        """
        Is true when ``argument`` is a metronome mark with quarters per
        minute greater than that of this metronome mark.
        """
        assert isinstance(argument, type(self)), repr(argument)
        self_quarters_per_minute = self.quarters_per_minute or 0
        argument_quarters_per_minute = argument.quarters_per_minute or 0
        assert isinstance(self_quarters_per_minute, (int, float, Fraction))
        assert isinstance(argument_quarters_per_minute, (int, float, Fraction))
        return self_quarters_per_minute < argument_quarters_per_minute

    def __mul__(self, multiplier) -> typing.Optional['MetronomeMark']:
        """
        Multiplies metronome mark by ``multiplier``.

        ..  container:: example

            Doubles metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> 2 * mark
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=168)

        ..  container:: example

            Triples metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> 3 * mark
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=252)

        """
        if not isinstance(multiplier, (int, float, Duration)):
            return None
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.units_per_minute, (int, float, Fraction))
        new_units_per_minute = multiplier * self.units_per_minute
        new_reference_duration = Duration(self.reference_duration)
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
            )
        return metronome_mark

    def __radd__(self, argument) -> typing.Optional['MetronomeMark']:
        """
        Adds ``argument`` to metronome mark.

        ..  container:: example

            Adds one metronome mark to another:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

            >>> mark_2 + mark_1
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=150)

        ..  container:: example

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example

            Raises imprecise metronome mark error with range:

            >>> mark_1 = abjad.MetronomeMark((1, 8), (90, 92))
            >>> mark_2 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_1 + mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        ..  container:: example

            Raises type error when ``argument`` is not a metronome mark:

            >>> 90 + abjad.MetronomeMark((1, 4), 60)
            Traceback (most recent call last):
                ...
            TypeError: 90

        """
        if not isinstance(argument, type(self)):
            raise TypeError(argument)
        return argument.__add__(self)

    def __rmul__(self, multiplier) -> typing.Optional['MetronomeMark']:
        """
        Multiplies ``multiplier`` by metronome mark.

        ..  container::: example

            Doubles metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> mark * 2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=168)

        ..  container::: example

            Triples metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 84)
            >>> mark * 3
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=252)

        """
        if not isinstance(multiplier, (int, float, Duration)):
            return None
        if self.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.units_per_minute, (int, float, Fraction))
        new_units_per_minute = multiplier * self.units_per_minute
        new_reference_duration = Duration(self.reference_duration)
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
            )
        return metronome_mark

    def __str__(self) -> str:
        """
        Gets string representation of metronome mark.

        ..  container:: example

            Integer-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> str(mark)
            '4=90'

        ..  container:: example

            Rational-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), (90, 96))
            >>> str(mark)
            '4=90-96'

        """
        if self.textual_indication is not None:
            string = self.textual_indication
        elif isinstance(self.units_per_minute, (int, float)):
            string = f'{self._dotted}={self.units_per_minute}'
        elif (isinstance(self.units_per_minute, Fraction) and
            not mathtools.is_integer_equivalent_number(self.units_per_minute)):
            integer_part = int(float(self.units_per_minute))
            remainder = self.units_per_minute - integer_part
            remainder = Fraction(remainder)
            string = f'{self._dotted}={integer_part}+{remainder}'
        elif (isinstance(self.units_per_minute, Fraction) and
            mathtools.is_integer_equivalent_number(self.units_per_minute)):
            string = '{}={}'
            integer = int(float(self.units_per_minute))
            string = string.format(self._dotted, integer)
        elif isinstance(self.units_per_minute, tuple):
            string = '{}={}-{}'
            string = string.format(
                self._dotted,
                self.units_per_minute[0],
                self.units_per_minute[1],
                )
        else:
            raise TypeError(f'unknown: {self.units_per_minute!r}.')
        return string

    def __sub__(self, argument) -> 'MetronomeMark':
        """
        Subtracts ``argument`` from metronome mark.

        ..  container:: example

            Same reference reference durations:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_2 = abjad.MetronomeMark((1, 4), 60)
            >>> mark_1 - mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Different reference durations:

            >>> mark_1 = abjad.MetronomeMark((1, 4), 90)
            >>> mark_2 = abjad.MetronomeMark((1, 2), 90)
            >>> mark_1 - mark_2
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=45)

        ..  container:: example

            Raises imprecise metronome mark error with textual indication:

            >>> mark_1 = abjad.MetronomeMark(textual_indication='Langsam')
            >>> mark_2 = abjad.MetronomeMark((1, 2), 90)
            >>> mark_1 - mark_2
            Traceback (most recent call last):
                ...
            abjad.exceptions.ImpreciseMetronomeMarkError

        """
        if not isinstance(argument, type(self)):
            raise Exception('must be metronome mark: {argument!r}.')
        if self.is_imprecise or argument.is_imprecise:
            raise exceptions.ImpreciseMetronomeMarkError
        assert isinstance(self.quarters_per_minute, (int, float, Fraction))
        assert isinstance(argument.quarters_per_minute, (int, float, Fraction))
        assert isinstance(self.reference_duration, Duration)
        assert isinstance(argument.reference_duration, Duration)
        new_quarters_per_minute = self.quarters_per_minute - \
            argument.quarters_per_minute
        minimum_denominator = min((
            self.reference_duration.denominator,
            argument.reference_duration.denominator,
            ))
        nonreduced_fraction = NonreducedFraction(new_quarters_per_minute / 4)
        nonreduced_fraction = nonreduced_fraction.with_denominator(
            minimum_denominator)
        new_units_per_minute, new_reference_duration_denominator = \
            nonreduced_fraction.pair
        new_reference_duration = Duration(
            1,
            new_reference_duration_denominator,
            )
        metronome_mark = type(self)(
            reference_duration=new_reference_duration,
            units_per_minute=new_units_per_minute,
            )
        return metronome_mark

    def __truediv__(self, argument) -> 'MetronomeMark':
        """
        Divides metronome mark by ``argument``.

        Operator required by Python 3.

        ..  container:: example

            Divides metronome mark by number:

            >>> abjad.MetronomeMark((1, 4), 60).__truediv__(2)
            MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=30)

        ..  container:: example

            Divides metronome mark by other metronome mark:

            >>> abjad.MetronomeMark((1, 4), 60).__truediv__(
            ...     abjad.MetronomeMark((1, 4), 40)
            ...     )
            Multiplier(3, 2)

        """
        return self.__div__(argument)

    ### PRIVATE PROPERTIES ###

    @property
    def _dotted(self):
        return self.reference_duration.lilypond_duration_string

    @property
    def _equation(self):
        if self.reference_duration is None:
            return
        if isinstance(self.units_per_minute, tuple):
            string = '{}={}-{}'
            string = string.format(
                self._dotted,
                self.units_per_minute[0],
                self.units_per_minute[1],
                )
            return string
        elif isinstance(self.units_per_minute, (float, Fraction)):
            markup = MetronomeMark.make_tempo_equation_markup(
                self.reference_duration,
                self.units_per_minute,
                )
            string = str(markup)
            return string
        string = f'{self._dotted}={self.units_per_minute}'
        return string

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        text, equation = None, None
        if self.textual_indication is not None:
            text = self.textual_indication
            text = Scheme.format_scheme_value(text)
        if (self.reference_duration is not None and
            self.units_per_minute is not None):
            equation = self._equation
        if self.custom_markup is not None:
            return rf'\tempo {self.custom_markup}'
        elif text and equation:
            return rf'\tempo {text} {equation}'
        elif equation:
            return rf'\tempo {equation}'
        elif text:
            return rf'\tempo {text}'
        else:
            return r'\tempo \default'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.hide:
            bundle.before.commands.append(self._get_lilypond_format())
        return bundle

    def _make_lhs_score_markup(self, reference_duration=None):
        import abjad
        reference_duration = reference_duration or self.reference_duration
        maker = abjad.NoteMaker()
        selection = maker([0], [reference_duration])
        markup = Duration._to_score_markup(selection)
        return markup

    def _get_markup(self, stem_height=1):
        if self.custom_markup is not None:
            return self.custom_markup
        duration_log = int(math.log(self.reference_duration.denominator, 2))
        markup = Markup.abjad_metronome_mark(
            duration_log,
            self.reference_duration.dot_count,
            stem_height,
            self.units_per_minute,
            )
        return markup

    ### PUBLIC PROPERTIES ###

    @property
    def custom_markup(self) -> typing.Optional[Markup]:
        r"""
        Gets custom markup of metronome mark.

        ..  container:: example

            With custom markup:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     abjad.Duration(1, 4),
            ...     67.5,
            ...     )
            >>> markup = markup.with_color('red')
            >>> mark = abjad.MetronomeMark(
            ...     reference_duration=(1, 4),
            ...     units_per_minute=Fraction(135, 2),
            ...     custom_markup=markup,
            ...     )
            >>> staff = abjad.Staff("c'4 d'4 e'4 f'4")
            >>> score = abjad.Score([staff])
            >>> abjad.attach(mark, staff[0])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(score)
                \new Score
                <<
                    \new Staff
                    {
                        \tempo \markup {
                            \with-color
                                #red
                                {
                                    \scale
                                        #'(0.75 . 0.75)
                                        \score
                                            {
                                                \new Score
                                                \with
                                                {
                                                    \override SpacingSpanner.spacing-increment = #0.5
                                                    proportionalNotationDuration = ##f
                                                }
                                                <<
                                                    \new RhythmicStaff
                                                    \with
                                                    {
                                                        \remove Time_signature_engraver
                                                        \remove Staff_symbol_engraver
                                                        \override Stem.direction = #up
                                                        \override Stem.length = #5
                                                        \override TupletBracket.bracket-visibility = ##t
                                                        \override TupletBracket.direction = #up
                                                        \override TupletBracket.minimum-length = #4
                                                        \override TupletBracket.padding = #1.25
                                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                                        \override TupletNumber.font-size = #0
                                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                                        tupletFullLength = ##t
                                                    }
                                                    {
                                                        c'4
                                                    }
                                                >>
                                                \layout {
                                                    indent = #0
                                                    ragged-right = ##t
                                                }
                                            }
                                    =
                                    \general-align
                                        #Y
                                        #-0.5
                                        67.5
                                }
                            }
                        c'4
                        d'4
                        e'4
                        f'4
                    }
                >>

        """
        return self._custom_markup

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            Fifty-two eighth notes per minute:

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.context
            'Score'

        ..  container:: example

            Ninety quarter notes per minute:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.context
            'Score'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def hide(self) -> typing.Optional[bool]:
        r"""
        Is true when metronome mark should not appear in output (but
        should still determine effective metronome mark).

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> metronome_mark_1 = abjad.MetronomeMark((1, 4), 72)
            >>> abjad.attach(metronome_mark_1, staff[0]) 
            >>> metronome_mark_2 = abjad.MetronomeMark(
            ...     textual_indication='Allegro',
            ...     hide=True,
            ...     )
            >>> abjad.attach(metronome_mark_2, staff[2]) 
            >>> score = abjad.Score([staff])
            >>> abjad.show(score) # doctest: +SKIP

            >>> abjad.f(score)
            \new Score
            <<
                \new Staff
                {
                    \tempo 4=72
                    c'4
                    d'4
                    e'4
                    f'4
                }
            >>

            >>> for leaf in abjad.iterate(staff).leaves():
            ...     prototype = abjad.MetronomeMark
            ...     leaf, abjad.inspect(leaf).effective(prototype)
            ...
            (Note("c'4"), MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=72))
            (Note("d'4"), MetronomeMark(reference_duration=Duration(1, 4), units_per_minute=72))
            (Note("e'4"), MetronomeMark(textual_indication='Allegro', hide=True))
            (Note("f'4"), MetronomeMark(textual_indication='Allegro', hide=True))

        """
        return self._hide

    @property
    def is_imprecise(self) -> bool:
        """
        Is true if metronome mark is entirely textual or if metronome
        mark's units_per_minute is a range.

        ..  container:: example

            Imprecise metronome marks:

            >>> abjad.MetronomeMark((1, 4), 60).is_imprecise
            False
            >>> abjad.MetronomeMark(4, 60, 'Langsam').is_imprecise
            False
            >>> abjad.MetronomeMark(textual_indication='Langsam').is_imprecise
            True
            >>> abjad.MetronomeMark(4, (35, 50), 'Langsam').is_imprecise
            True
            >>> abjad.MetronomeMark((1, 4), (35, 50)).is_imprecise
            True

        ..  container:: example

            Precise metronome marks:

            >>> abjad.MetronomeMark((1, 4), 60).is_imprecise
            False

        """
        if self.reference_duration is not None:
            if self.units_per_minute is not None:
                if not isinstance(self.units_per_minute, tuple):
                    return False
        return True

    @property
    def parameter(self) -> str:
        """
        Is ``'METRONOME_MARK'``.

        ..  container:: example

            >>> abjad.MetronomeMark((1, 8), 52).parameter
            'METRONOME_MARK'

        """
        return self._parameter

    @property
    def quarters_per_minute(self) -> typing.Union[tuple, None, Fraction]:
        """
        Gets metronome mark quarters per minute.

        ..  container:: example

            Fifty-two eighth notes per minute:

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.quarters_per_minute
            Fraction(104, 1)

        ..  container:: example

            Ninety quarter notes per minute:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.quarters_per_minute
            Fraction(90, 1)

        ..  container:: example

            140 quarters per minute:

            >>> mark = abjad.MetronomeMark((3, 32), Fraction(105, 2))
            >>> mark.quarters_per_minute
            Fraction(140, 1)

        Returns tuple when metronome mark ``units_per_minute`` is a range.

        Returns none when metronome mark is imprecise.

        Returns fraction otherwise.
        """
        if self.is_imprecise:
            return None
        if isinstance(self.units_per_minute, tuple):
            low = Duration(1, 4) / self.reference_duration * \
                self.units_per_minute[0]
            high = Duration(1, 4) / self.reference_duration * \
                self.units_per_minute[1]
            return (low, high)
        result = Duration(1, 4) / self.reference_duration * \
            self.units_per_minute
        return Fraction(result)

    @property
    def reference_duration(self) -> typing.Optional[Duration]:
        """
        Gets reference duration of metronome mark.

        ..  container:: example

            Fifty-two eighth notes per minute:

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.reference_duration
            Duration(1, 8)

        ..  container:: example

            Ninety quarter notes per minute:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.reference_duration
            Duration(1, 4)

        """
        return self._reference_duration

    @property
    def textual_indication(self) -> typing.Optional[str]:
        """
        Gets optional textual indication of metronome mark.

        ..  container:: example

            Fifty-two eighth notes per minute:

            >>> mark = abjad.MetronomeMark((1, 8), 52)
            >>> mark.textual_indication is None
            True

        ..  container:: example

            Ninety quarter notes per minute:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.textual_indication is None
            True

        """
        return self._textual_indication

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on metronome mark.
        
        The LilyPond ``\tempo`` command refuses tweaks.

        Override the LilyPond ``MetronomeMark`` grob instead.
        """
        pass

    @property
    def units_per_minute(self) -> typing.Union[int, Fraction, None]:
        """
        Gets units per minute of metronome mark.

        ..  container:: example

            Integer-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), 90)
            >>> mark.units_per_minute
            90

        ..  container:: example

            Rational-valued metronome mark:

            >>> mark = abjad.MetronomeMark((1, 4), abjad.Fraction(181, 2))
            >>> mark.units_per_minute
            Fraction(181, 2)

        """
        return self._units_per_minute

    ### PUBLIC METHODS ###

    def duration_to_milliseconds(self, duration) -> Duration:
        """
        Gets millisecond value of ``duration`` under a given metronome mark.

        ..  container:: example

            One quarter lasts 1000 msec at quarter equals 60:

            >>> mark = abjad.MetronomeMark((1, 4), 60)
            >>> mark.duration_to_milliseconds((1, 4))
            Duration(1000, 1)

        ..  container:: example

            Dotted sixteenth lasts 1500 msec at quarter equals 60:

            >>> mark = abjad.MetronomeMark((1, 4), 60)
            >>> mark.duration_to_milliseconds((3, 8))
            Duration(1500, 1)

        """
        assert isinstance(self.reference_duration, Duration)
        denominator = self.reference_duration.denominator
        numerator = self.reference_duration.numerator
        whole_note_duration = 1000
        whole_note_duration *= Multiplier(denominator, numerator)
        whole_note_duration *= Multiplier(60, self.units_per_minute)
        duration = Duration(duration)
        return Duration(duration * whole_note_duration)

    def list_related_tempos(
        self,
        maximum_numerator=None,
        maximum_denominator=None,
        integer_tempos_only=False,
        ):
        r"""
        Lists related tempos.

        ..  container:: example

            Rewrites tempo ``4=58`` by ratios ``n:d`` such that ``1 <= n <= 8``
            and ``1 <= d <= 8``.

            >>> mark = abjad.MetronomeMark((1, 4), 58)
            >>> pairs = mark.list_related_tempos(
            ...     maximum_numerator=8,
            ...     maximum_denominator=8,
            ...     )

            >>> for tempo, ratio in pairs:
            ...     string = f'{tempo!s}\t{ratio!s}'
            ...     print(string)
            4=29        1:2
            4=33+1/7    4:7
            4=34+4/5    3:5
            4=36+1/4    5:8
            4=38+2/3    2:3
            4=41+3/7    5:7
            4=43+1/2    3:4
            4=46+2/5    4:5
            4=48+1/3    5:6
            4=49+5/7    6:7
            4=50+3/4    7:8
            4=58        1:1
            4=66+2/7    8:7
            4=67+2/3    7:6
            4=69+3/5    6:5
            4=72+1/2    5:4
            4=77+1/3    4:3
            4=81+1/5    7:5
            4=87        3:2
            4=92+4/5    8:5
            4=96+2/3    5:3
            4=101+1/2   7:4
            4=116       2:1

        ..  container:: example

            Integer-valued tempos only:

            >>> mark = abjad.MetronomeMark((1, 4), 58)
            >>> pairs = mark.list_related_tempos(
            ...     maximum_numerator=16,
            ...     maximum_denominator=16,
            ...     integer_tempos_only=True,
            ...     )

            >>> for tempo, ratio in pairs:
            ...     string = f'{tempo!s}\t{ratio!s}'
            ...     print(string)
            4=29	1:2
            4=58	1:1
            4=87	3:2
            4=116	2:1

        Constrains ratios such that ``1:2 <= n:d <= 2:1``.

        Returns list of tempo / ratio pairs.
        """
        allowable_numerators = range(1, maximum_numerator + 1)
        allowable_denominators = range(1, maximum_denominator + 1)
        numbers = [allowable_numerators, allowable_denominators]
        enumerator = Enumerator(numbers)
        pairs = enumerator.yield_outer_product()
        multipliers = [Multiplier(_) for _ in pairs]
        multipliers = [
            _ for _ in multipliers
            if Fraction(1, 2) <= _ <= Fraction(2)
            ]
        multipliers.sort()
        multipliers = sequence(multipliers).remove_repeats()
        pairs = []
        for multiplier in multipliers:
            new_units_per_minute = multiplier * self.units_per_minute
            if (integer_tempos_only and not
                mathtools.is_integer_equivalent_number(new_units_per_minute)):
                continue
            metronome_mark = type(self)(
                reference_duration=self.reference_duration,
                units_per_minute=new_units_per_minute,
                )
            ratio = Ratio(multiplier.pair)
            pair = (metronome_mark, ratio)
            pairs.append(pair)
        return pairs

    @staticmethod
    def make_tempo_equation_markup(reference_duration, units_per_minute):
        r"""
        Makes tempo equation markup.

        ..  container:: example

            Integer-valued metronome mark:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     (1, 4),
            ...     90,
            ...     )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \general-align
                        #Y
                        #-0.5
                        90
                    }

        ..  container:: example

            Float-valued metronome mark:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     (1, 4),
            ...     90.1,
            ...     )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \general-align
                        #Y
                        #-0.5
                        90.1
                    }

        ..  container:: example

            Rational-valued metronome mark:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     abjad.Duration(1, 4),
            ...     abjad.Fraction(181, 2),
            ...     )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \raise
                        #-0.5
                        {
                            90
                            \tiny
                                \fraction
                                    1
                                    2
                        }
                    }

        ..  container:: example

            Reference duration expressed with ties:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(
            ...     (5, 16),
            ...     90,
            ...     )
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'4
                                        ~
                                        c'16
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \general-align
                        #Y
                        #-0.5
                        90
                    }

        ..  container:: example

            Reference duration expressed as a tuplet:

            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup((1, 6), 90)
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        \tweak edge-height #'(0.7 . 0)
                                        \times 2/3 {
                                            c'4
                                        }
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \general-align
                        #Y
                        #-0.5
                        90
                    }

        ..  container:: example

            Reference duration passed in as explicit rhythm:

            >>> maker = abjad.NoteMaker()
            >>> durations = [(1, 16), (3, 16), (1, 16)]
            >>> selection = maker([0], durations)
            >>> abjad.attach(abjad.Tie(), selection)
            >>> abjad.attach(abjad.Beam(), selection)
            >>> markup = abjad.MetronomeMark.make_tempo_equation_markup(selection, 90)
            >>> abjad.show(markup) # doctest: +SKIP

            ..  docs::

                >>> print(format(markup))
                \markup {
                    \scale
                        #'(0.75 . 0.75)
                        \score
                            {
                                \new Score
                                \with
                                {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                }
                                <<
                                    \new RhythmicStaff
                                    \with
                                    {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.minimum-length = #4
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                                        \override TupletNumber.font-size = #0
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    }
                                    {
                                        c'16
                                        ~
                                        [
                                        c'8.
                                        ~
                                        c'16
                                        ]
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                    =
                    \general-align
                        #Y
                        #-0.5
                        90
                    }

            Pass rhythms like this as Abjad selections.

        Returns markup.
        """
        import abjad
        if isinstance(reference_duration, abjad.Selection):
            selection = reference_duration
        else:
            maker = abjad.NoteMaker()
            selection = maker([0], [reference_duration])
        lhs_score_markup = Duration._to_score_markup(selection)
        lhs_score_markup = lhs_score_markup.scale((0.75, 0.75))
        equal_markup = abjad.Markup('=')
        if (isinstance(units_per_minute, abjad.Fraction) and
            not abjad.mathtools.is_integer_equivalent_number(units_per_minute)):
            rhs_markup = abjad.Markup.make_improper_fraction_markup(
                units_per_minute)
            rhs_markup = rhs_markup.raise_(-0.5)
        else:
            rhs_markup = abjad.Markup(units_per_minute)
            rhs_markup = rhs_markup.general_align('Y', -0.5)
        markup = lhs_score_markup + equal_markup + rhs_markup
        return markup

    def rewrite_duration(self, duration, metronome_mark):
        r"""
        Rewrites ``duration`` under ``metronome_mark``.

        ..  container:: example

            Consider the two metronome marks below.

            >>> tempo = abjad.MetronomeMark((1, 4), 60)
            >>> metronome_mark = abjad.MetronomeMark((1, 4), 90)

            ``tempo`` specifies quarter equal to ``60``.

            ``metronome_mark`` indication specifies quarter equal to ``90``.

            ``metronome_mark`` is ``3/2`` times as fast as ``tempo``:

            >>> metronome_mark / tempo
            Multiplier(3, 2)

            Note that a triplet eighth note under ``tempo`` equals a regular
            eighth note under ``metronome_mark``:

            >>> tempo.rewrite_duration((1, 12), metronome_mark)
            Duration(1, 8)

            And note that a regular eighth note under ``tempo`` equals a dotted
            sixteenth under ``metronome_mark``:

            >>> tempo.rewrite_duration((1, 8), metronome_mark)
            Duration(3, 16)

        Given ``duration`` governed by this tempo returns new duration governed
        by ``metronome_mark``.

        Ensures that ``duration`` and new duration consume the same amount of
        time in seconds.

        Returns duration.
        """
        import abjad
        duration = Duration(duration)
        tempo_ratio = metronome_mark / self
        new_duration = tempo_ratio * duration
        return new_duration



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.StorageFormatManager import StorageFormatManager


class Tremolo(AbjadValueObject):
    r"""
    Tremolo (of exactly two notes).

    ..  container:: example

        With two beams:

        >>> chord = abjad.Chord("<cs' e'>4")
        >>> tremolo = abjad.Tremolo(beam_count=2)
        >>> abjad.attach(tremolo, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(chord)
            \repeat tremolo 2
            {
            cs'16 e'16
            }

    ..  container:: example

        With three beams:

        >>> chord = abjad.Chord("<cs' e'>4")
        >>> tremolo = abjad.Tremolo(beam_count=3)
        >>> abjad.attach(tremolo, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(chord)
            \repeat tremolo 4
            {
            cs'32 e'32
            }

    Tremolo affects the formatting of chords.

    Tremolo has no effect when attached to notes or rests.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_beam_count',
        '_is_slurred',
        )

    _format_slot = None

    ### INITIALIZER ###

    def __init__(
        self,
        beam_count: int = 3,
        *,
        is_slurred: bool = None,
        ) -> None:
        assert isinstance(beam_count, int), repr(beam_count)
        assert 0 < beam_count, repr(beam_count)
        self._beam_count = beam_count
        if is_slurred is not None:
            is_slurred = bool(is_slurred)
        self._is_slurred = is_slurred

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments) -> 'Tremolo':
        """
        Copies tremolo.

        ..  container:: example

            Copies tremolo:

            >>> import copy
            >>> tremolo_1 = abjad.Tremolo(beam_count=2)
            >>> tremolo_2 = copy.copy(tremolo_1)

            >>> tremolo_1 == tremolo_2
            True

            >>> tremolo_1 is not tremolo_2
            True

        """
        return super().__copy__(*arguments)

    def __format__(self, format_specification='') -> str:
        """
        Formats stem tremolo.

        ..  container:: example

            With two beams:

            >>> tremolo = abjad.Tremolo(beam_count=2)
            >>> print(format(tremolo))
            abjad.Tremolo(
                beam_count=2,
                )

        ..  container:: example

            With three beams:

            >>> tremolo = abjad.Tremolo(beam_count=3)
            >>> print(format(tremolo))
            abjad.Tremolo(
                beam_count=3,
                )

        """
        if format_specification in ('', 'storage'):
            return StorageFormatManager(self).get_storage_format()
        assert format_specification == 'lilypond'
        raise Exception('no LilyPond format available.')

    def __str__(self) -> str:
        """
        Gets string representation of tremolo.

        ..  container:: example

            >>> tremolo = abjad.Tremolo(beam_count=2)
            >>> str(tremolo)
            'Tremolo(beam_count=2)'

            >>> tremolo = abjad.Tremolo(beam_count=3)
            >>> str(tremolo)
            'Tremolo(beam_count=3)'

        """
        return super().__str__()

    ### PUBLIC PROPERTIES ###

    @property
    def beam_count(self) -> int:
        r"""
        Gets beam count of tremolo.

        ..  container:: example

            With two beams:

            >>> chord = abjad.Chord("<cs' e'>4")
            >>> tremolo = abjad.Tremolo(beam_count=2)
            >>> abjad.attach(tremolo, chord)
            >>> abjad.show(chord) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(chord)
                \repeat tremolo 2
                {
                cs'16 e'16
                }

        ..  container:: example

            With three beams:

            >>> chord = abjad.Chord("<cs' e'>4")
            >>> tremolo = abjad.Tremolo()
            >>> abjad.attach(tremolo, chord)
            >>> abjad.show(chord) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(chord)
                \repeat tremolo 4
                {
                cs'32 e'32
                }

        """
        return self._beam_count

    @property
    def is_slurred(self) -> typing.Optional[bool]:
        r"""
        Is true when tremolo is slurred.

        ..  container:: example

            Without slur:

            >>> chord = abjad.Chord("<cs' e'>4")
            >>> tremolo = abjad.Tremolo()
            >>> abjad.attach(tremolo, chord)
            >>> abjad.show(chord) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(chord)
                \repeat tremolo 4
                {
                cs'32 e'32
                }

        ..  container:: example

            With slur:

            >>> chord = abjad.Chord("<cs' e'>4")
            >>> tremolo = abjad.Tremolo(is_slurred=True)
            >>> abjad.attach(tremolo, chord)
            >>> abjad.show(chord) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(chord)
                \repeat tremolo 4
                {
                cs'32 \( e'32 \)
                }

        """
        return self._is_slurred

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on tremolo.
        
        The LilyPond ``\repeat tremolo`` command refuses tweaks.

        Override the LilyPond ``Beam`` grob instead.
        """
        pass



from .Duration import Duration


class Offset(Duration):
    """
    Offset.

    ..  container:: example

        Initializes from integer numerator:

        >>> abjad.Offset(3)
        Offset(3, 1)

    ..  container:: example

        Initializes from integer numerator and denominator:

        >>> abjad.Offset(3, 16)
        Offset(3, 16)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator:

        >>> abjad.Offset(3.0)
        Offset(3, 1)

    ..  container:: example

        Initializes from integer-equivalent numeric numerator and denominator:

        >>> abjad.Offset(3.0, 16)
        Offset(3, 16)

    ..  container:: example

        Initializes from integer-equivalent singleton:

        >>> abjad.Offset((3,))
        Offset(3, 1)

    ..  container:: example

        Initializes from integer-equivalent pair:

        >>> abjad.Offset((3, 16))
        Offset(3, 16)

    ..  container:: example

        Initializes from duration:

        >>> abjad.Offset(abjad.Duration(3, 16))
        Offset(3, 16)

    ..  container:: example

        Initializes from other offset:

        >>> abjad.Offset(abjad.Offset(3, 16))
        Offset(3, 16)

    ..  container:: example

        Initializes from other offset with grace displacement:

        >>> offset = abjad.Offset((3, 16), grace_displacement=(-1, 16))
        >>> abjad.Offset(offset)
        Offset(
            (3, 16),
            grace_displacement=Duration(-1, 16)
            )

    ..  container:: example

        Intializes from fraction:

        >>> abjad.Offset(abjad.Fraction(3, 16))
        Offset(3, 16)

    ..  container:: example

        Initializes from solidus string:

        >>> abjad.Offset('3/16')
        Offset(3, 16)

    ..  container:: example

        Initializes from nonreduced fraction:

        >>> abjad.Offset(abjad.NonreducedFraction(6, 32))
        Offset(3, 16)

    ..  container:: example

        Offsets inherit from built-in fraction:

        >>> isinstance(abjad.Offset(3, 16), abjad.Fraction)
        True

    ..  container:: example

        Offsets are numbers:

        >>> import numbers

        >>> isinstance(abjad.Offset(3, 16), numbers.Number)
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_grace_displacement',
        )

    ### CONSTRUCTOR ###

    def __new__(class_, *arguments, **keywords):
        grace_displacement = None
        for argument in arguments:
            try:
                grace_displacement = argument.grace_displacement
                break
            except AttributeError:
                pass
        grace_displacement = grace_displacement or keywords.get(
            'grace_displacement')
        if grace_displacement is not None:
            grace_displacement = Duration(grace_displacement)
        grace_displacement = grace_displacement or None
        if len(arguments) == 1 and isinstance(arguments[0], Duration):
            arguments = arguments[0].pair
        self = Duration.__new__(class_, *arguments)
        self._grace_displacement = grace_displacement
        return self

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments):
        """
        Copies offset.

        >>> import copy

        ..  container:: example

            Copies offset with grace displacement:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = copy.copy(offset_1)

            >>> offset_1
            Offset(
                (1, 4),
                grace_displacement=Duration(-1, 16)
                )

            >>> offset_2
            Offset(
                (1, 4),
                grace_displacement=Duration(-1, 16)
                )

            >>> offset_1 == offset_2
            True

            >>> offset_1 is offset_2
            False

        Returns new offset.
        """
        return type(self)(
            self.pair,
            grace_displacement=self.grace_displacement
            )

    def __deepcopy__(self, *arguments):
        """
        Deep copies offset.

        >>> import copy

        ..  container:: example

            Copies offset with grace displacement:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = copy.deepcopy(offset_1)

            >>> offset_1
            Offset(
                (1, 4),
                grace_displacement=Duration(-1, 16)
                )

            >>> offset_2
            Offset(
                (1, 4),
                grace_displacement=Duration(-1, 16)
                )

            >>> offset_1 == offset_2
            True

            >>> offset_1 is offset_2
            False

        Returns new offset.
        """
        return self.__copy__(*arguments)

    def __eq__(self, argument):
        """
        Is true when offset equals ``argument``.

        ..  container:: example

            With equal numerators, denominators and grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            True
            >>> offset_2 == offset_1
            True
            >>> offset_2 == offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            False
            >>> offset_2 == offset_1
            False
            >>> offset_2 == offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), grace_displacement=(-99))

            >>> offset_1 == offset_1
            True
            >>> offset_1 == offset_2
            False
            >>> offset_2 == offset_1
            False
            >>> offset_2 == offset_2
            True

        Returns true or false.
        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_grace_displacement() == \
                argument._get_grace_displacement()
        return super().__eq__(argument)

    def __ge__(self, argument):
        """
        Is true when offset is greater than or equal to ``argument``.

        ..  container:: example

            With equal numerators, denominators and grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            True
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            False
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), grace_displacement=(-99))

            >>> offset_1 >= offset_1
            True
            >>> offset_1 >= offset_2
            False
            >>> offset_2 >= offset_1
            True
            >>> offset_2 >= offset_2
            True

        Returns true or false.
        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_grace_displacement() >= \
                argument._get_grace_displacement()
        return super().__ge__(argument)

    def __gt__(self, argument):
        """
        Is true when offset is greater than ``argument``.

        ..  container:: example

            With equal numerators, denominators and grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            False
            >>> offset_2 > offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            True
            >>> offset_2 > offset_2
            False

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), grace_displacement=(-99))

            >>> offset_1 > offset_1
            False
            >>> offset_1 > offset_2
            False
            >>> offset_2 > offset_1
            True
            >>> offset_2 > offset_2
            False

        Returns true or false.
        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_grace_displacement() > \
                argument._get_grace_displacement()
        return Duration.__gt__(self, argument)

    def __hash__(self):
        """
        Hashes offset.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __le__(self, argument):
        """
        Is true when offset is less than or equal to ``argument``.

        ..  container:: example

            With equal numerators, denominators and grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            True
            >>> offset_2 <= offset_2
            True

        ..  container:: example

            With equal numerators and denominators but differing grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            False
            >>> offset_2 <= offset_2
            True

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), grace_displacement=(-99))

            >>> offset_1 <= offset_1
            True
            >>> offset_1 <= offset_2
            True
            >>> offset_2 <= offset_1
            False
            >>> offset_2 <= offset_2
            True

        Returns true or false.
        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_grace_displacement() <= \
                argument._get_grace_displacement()
        return super().__le__(argument)

    def __lt__(self, argument):
        """
        Is true when offset is less than ``argument``.

        ..  container:: example

            With equal numerators, denominators and grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 16))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            False
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing nonzero grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4), grace_displacement=(-1, 16))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With equal numerators and denominators but differing zero-valued
            grace displacements:

            >>> offset_1 = abjad.Offset((1, 4), grace_displacement=(-1, 8))
            >>> offset_2 = abjad.Offset((1, 4))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        ..  container:: example

            With differing numerators and denominators. Ignores grace
            displacements:

            >>> offset_1 = abjad.Offset((1, 4))
            >>> offset_2 = abjad.Offset((1, 2), grace_displacement=(-99))

            >>> offset_1 < offset_1
            False
            >>> offset_1 < offset_2
            True
            >>> offset_2 < offset_1
            False
            >>> offset_2 < offset_2
            False

        Returns true or false.
        """
        if isinstance(argument, type(self)) and self.pair == argument.pair:
            return self._get_grace_displacement() < \
                argument._get_grace_displacement()
        return super().__lt__(argument)

    def __repr__(self):
        """
        Gets interpreter representation of offset.

        ..  container:: example

            Gets interpreter representation of offset without grace
            displacement:

            >>> abjad.Offset(1, 4)
            Offset(1, 4)

        ..  container:: example

            Gets interpreter representation of offset with grace displacement:

            >>> abjad.Offset(1, 4, grace_displacement=(-1, 16))
            Offset(
                (1, 4),
                grace_displacement=Duration(-1, 16)
                )

        """
        return super().__repr__()

    def __sub__(self, argument):
        """
        Offset taken from offset returns duration:

        >>> abjad.Offset(2) - abjad.Offset(1, 2)
        Duration(3, 2)

        Duration taken from offset returns another offset:

        >>> abjad.Offset(2) - abjad.Duration(1, 2)
        Offset(3, 2)

        Coerce ``argument`` to offset when ``argument`` is neither offset nor
        duration:

        >>> abjad.Offset(2) - abjad.Fraction(1, 2)
        Duration(3, 2)

        Returns duration or offset.
        """
        if isinstance(argument, type(self)):
            return Duration(super().__sub__(argument))
        elif isinstance(argument, Duration):
            return super().__sub__(argument)
        else:
            argument = type(self)(argument)
            return self - argument

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        from abjad.system.FormatSpecification import \
            FormatSpecification
        is_indented = False
        names = []
        values = [self.numerator, self.denominator]
        if self._get_grace_displacement():
            is_indented = True
            names = ['grace_displacement']
            values = [(self.numerator, self.denominator)]
        return FormatSpecification(
            client=self,
            repr_is_indented=is_indented,
            storage_format_args_values=values,
            storage_format_is_indented=is_indented,
            storage_format_kwargs_names=names,
            )

    def _get_grace_displacement(self):
        import abjad
        if self.grace_displacement is None:
            return abjad.Duration(0)
        return self.grace_displacement

    ### PUBLIC PROPERTIES ###

    @property
    def grace_displacement(self):
        """
        Gets grace displacement.

        ..  container:: example

            Gets grace displacement equal to none:

            >>> offset = abjad.Offset(1, 4)
            >>> offset.grace_displacement is None
            True

        ..  container:: example

            Gets grace displacement equal to a negative sixteenth:

            >>> offset = abjad.Offset(1, 4, grace_displacement=(-1, 16))
            >>> offset.grace_displacement
            Duration(-1, 16)

        ..  container:: example

            Stores zero-valued grace displacement as none:

            >>> offset = abjad.Offset(1, 4, grace_displacement=0)
            >>> offset.grace_displacement is None
            True

            >>> offset
            Offset(1, 4)

        Defaults to none.

        Set to duration or none.

        Returns duration or none.
        """
        return self._grace_displacement



"""
Tools for modeling LilyPond's "grobs", "context", "engravers", etc.
"""

from .LilyPondContext import LilyPondContext
from .LilyPondContextSetting import LilyPondContextSetting
from .LilyPondEngraver import LilyPondEngraver
from .LilyPondGrob import LilyPondGrob
from .LilyPondGrobInterface import LilyPondGrobInterface
from .LilyPondNameManager import LilyPondNameManager
from .LilyPondGrobNameManager import LilyPondGrobNameManager
from .LilyPondGrobOverride import LilyPondGrobOverride
from .LilyPondSettingNameManager import LilyPondSettingNameManager
from .LilyPondTweakManager import LilyPondTweakManager



def all_are_integer_equivalent_numbers(argument):
    """
    Is true when ``argument`` is an iterable collection with
    integer-equivalent items.

    ..  container:: example

        >>> items = [1, 2, 3.0, abjad.Fraction(4, 1)]
        >>> abjad.mathtools.all_are_integer_equivalent_numbers(items)
        True

        >>> abjad.mathtools.all_are_integer_equivalent_numbers([1, 2, 3.5, 4])
        False

    Returns true or false.
    """
    from abjad import mathtools
    try:
        return all(mathtools.is_integer_equivalent_number(_) for _ in argument)
    except TypeError:
        return False



from abjad.system.AbjadValueObject import AbjadValueObject
from .NumberedPitchClass import NumberedPitchClass


class ColorMap(AbjadValueObject):
    """
    Color map.

    ..  container:: example

        Maps pitch-classes to red, green and blue:

        >>> color_map = abjad.ColorMap(
        ...     colors=['red', 'green', 'blue'],
        ...     pitch_iterables=[
        ...         [-8, 2, 10, 21],
        ...         [0, 11, 32, 41],
        ...         [15, 25, 42, 43],
        ...         ],
        ...     )

        >>> abjad.f(color_map)
        abjad.ColorMap(
            colors=['red', 'green', 'blue'],
            pitch_iterables=[
                [-8, 2, 10, 21],
                [0, 11, 32, 41],
                [15, 25, 42, 43],
                ],
            )

    Color maps are immutable.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_color_dictionary',
        '_colors',
        '_pitch_iterables',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, *, colors=None, pitch_iterables=None):
        pitch_iterables = pitch_iterables or []
        colors = colors or []
        assert len(pitch_iterables) == len(colors)
        self._pitch_iterables = pitch_iterables
        self._colors = colors
        self._color_dictionary = {}
        self._initialize_color_dictionary()

    ### SPECIAL METHODS ###

    def __format__(self, format_specification=''):
        """
        Formats color map.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> abjad.f(color_map)
            abjad.ColorMap(
                colors=['red', 'green', 'blue'],
                pitch_iterables=[
                    [-8, 2, 10, 21],
                    [0, 11, 32, 41],
                    [15, 25, 42, 43],
                    ],
                )

        Returns string.
        """
        return super().__format__(format_specification=format_specification)

    def __getitem__(self, pitch_class):
        """
        Gets `pitch_class` color.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map[11]
            'green'

        Returns string.
        """
        pitch_class = NumberedPitchClass(pitch_class)
        return self._color_dictionary[pitch_class.number]

    ### PRIVATE METHODS ###

    def _initialize_color_dictionary(self):
        for pitch_iterable, color in zip(self.pitch_iterables, self.colors):
            for pitch in pitch_iterable:
                pc = NumberedPitchClass(pitch)
                keys = self._color_dictionary.keys()
                if pc.number in list(keys):
                    print(pc, list(self._color_dictionary.keys()))
                    message = 'duplicated pitch-class in color map: {!r}.'
                    message = message.format(pc)
                    raise KeyError(message)
                self._color_dictionary[pc.number] = color

    ### PUBLIC PROPERTIES ###

    @property
    def colors(self):
        """
        Gets colors.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map.colors
            ['red', 'green', 'blue']

        Returns list.
        """
        return self._colors

    @property
    def is_twelve_tone_complete(self):
        """
        Is true when color map contains all 12-ET pitch-classes.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map.is_twelve_tone_complete
            True

        Return boolean.
        """
        pcs = range(12)
        return set(pcs).issubset(set(self._color_dictionary.keys()))

    @property
    def is_twenty_four_tone_complete(self):
        """
        Is true when color map contains all 24-ET pitch-classes.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map.is_twenty_four_tone_complete
            False

        Return boolean.
        """
        pcs = [x / 2.0 for x in range(24)]
        pcs = [int(x) if int(x) == x else x for x in pcs]
        return set(pcs).issubset(set(self._color_dictionary.keys()))

    @property
    def pairs(self):
        """
        Gets pairs.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> for pair in color_map.pairs:
            ...     pair
            ...
            (0, 'green')
            (1, 'blue')
            (2, 'red')
            (3, 'blue')
            (4, 'red')
            (5, 'green')
            (6, 'blue')
            (7, 'blue')
            (8, 'green')
            (9, 'red')
            (10, 'red')
            (11, 'green')

        Returns list.
        """
        items = list(self._color_dictionary.items())
        return list(sorted(items))

    @property
    def pitch_iterables(self):
        """
        Gets pitch iterables.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map.pitch_iterables
            [[-8, 2, 10, 21], [0, 11, 32, 41], [15, 25, 42, 43]]

        Returns list.
        """
        return self._pitch_iterables

    ### PUBLIC METHODS ###

    def get(self, key, alternative=None):
        """
        Gets `key` from color map.

        ..  container:: example

            >>> color_map = abjad.ColorMap(
            ...     colors=['red', 'green', 'blue'],
            ...     pitch_iterables=[
            ...         [-8, 2, 10, 21],
            ...         [0, 11, 32, 41],
            ...         [15, 25, 42, 43],
            ...         ],
            ...     )

            >>> color_map.get(11)
            'green'

        Returns `alternative` when `key` is not found.

        Returns string.
        """
        try:
            return self[key]
        except (KeyError, TypeError, ValueError):
            return alternative



def f(argument, strict=None):
    r"""
    Formats ``argument`` and prints result.
    
    ..  container:: example

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup('Allegro', direction=abjad.Up)
        >>> markup = markup.with_color('blue')
        >>> abjad.attach(markup, staff[0])
        >>> for leaf in staff:
        ...     abjad.attach(abjad.Articulation('.'), leaf)
        ...
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            -\staccato
            ^ \markup {
                \with-color
                    #blue
                    Allegro
                }
            d'4
            -\staccato
            e'4
            -\staccato
            f'4
            -\staccato
        }

        >>> abjad.show(staff) # doctest: +SKIP

    """
    import abjad
    if strict is not None:
        assert isinstance(strict, int), repr(strict)
    if hasattr(argument, '_publish_storage_format'):
        string = format(argument, 'storage')
    else:
        string = format(argument, 'lilypond')
    realign = None
    if isinstance(strict, int):
        string = abjad.LilyPondFormatManager.align_tags(string, strict)
        realign = strict
    string = abjad.LilyPondFormatManager.left_shift_tags(
        string,
        realign=realign,
        )
    print(string)



import collections
import copy
import functools
import numbers
from . import constants
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.StorageFormatManager import StorageFormatManager


@functools.total_ordering
class PitchRange(AbjadValueObject):
    """
    Pitch range.

    ..  container:: example

        Pitches from C3 to C7, inclusive:

        >>> pitch_range = abjad.PitchRange('[C3, C7]')
        >>> abjad.show(pitch_range) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(pitch_range)
            abjad.PitchRange('[C3, C7]')

    Initalizes from pitch numbers, pitch names, pitch instances,
    one-line reprs or other pitch range objects.

    Pitch ranges do not sort relative to other pitch ranges.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_start',
        '_stop',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, range_string='[A0, C8]'):
        if isinstance(range_string, type(self)):
            range_string = range_string.range_string
        start, stop = self._parse_range_string(range_string)
        self._start = start
        self._stop = stop

    ### SPECIAL METHODS ###

    def __contains__(self, argument):
        """
        Is true when pitch range contains `argument`.

        ..  container:: example

            Closed / closed range:

            >>> range_ = abjad.PitchRange('[A0, C8]')

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Closed / open range:

            >>> range_ = abjad.PitchRange('[A0, C8)')

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Closed / infinite range:

            >>> range_ = abjad.PitchRange('[-39, +inf]')

            >>> -99 in range_
            False

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            True

        ..  container:: example

            Open / closed range:

            >>> range_ = abjad.PitchRange('(A0, C8]')

            >>> -99 in range_
            False

            >>> -39 in range_
            False

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Open / open range:

            >>> range_ = abjad.PitchRange('(A0, C8)')

            >>> -99 in range_
            False

            >>> -39 in range_
            False

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / closed range:

            >>> range_ = abjad.PitchRange('[-inf, C8]')

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / open range:

            >>> range_ = abjad.PitchRange('[-inf, C8)')

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            False

            >>> 99 in range_
            False

        ..  container:: example

            Infinite / infinite range:

            >>> range_ = abjad.PitchRange('[-inf, +inf]')

            >>> -99 in range_
            True

            >>> -39 in range_
            True

            >>> 0 in range_
            True

            >>> 48 in range_
            True

            >>> 99 in range_
            True

        Returns true or false.
        """
        import abjad
        if (
            hasattr(argument, '_has_effective_indicator') and
            'unpitched' in abjad.inspect(argument).indicators(str)
            ):
            return True
        elif isinstance(argument, (int, float)):
            pitch = abjad.NamedPitch(argument)
            return self._contains_pitch(pitch)
        elif isinstance(argument, abjad.NamedPitch):
            return self._contains_pitch(argument)
        elif isinstance(argument, abjad.Note):
            sounding_pitch = abjad.inspect(argument).sounding_pitch()
            return self._contains_pitch(sounding_pitch)
        elif isinstance(argument, abjad.Chord):
            sounding_pitches = abjad.inspect(argument).sounding_pitches()
            return all(self._contains_pitch(x) for x in sounding_pitches)
        elif isinstance(argument, (abjad.Rest, abjad.Skip)):
            return True
        elif isinstance(argument, abjad.Container):
            return all(x in self for x in abjad.iterate(argument).leaves())
        else:
            pitches = list(abjad.iterate(argument).pitches())
            if pitches:
                return all(self._contains_pitch(x) for x in pitches)
            else:
                try:
                    return all(self._contains_pitch(x) for x in argument)
                except TypeError:
                    return False
        return False

    def __eq__(self, argument):
        """
        Is true when `argument` is a pitch range with start and stop equal
        to those of this pitch range.

        ..  container:: example

            >>> range_1 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_2 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_3 = abjad.PitchRange.from_pitches(-39, 48)

            >>> range_1 == range_1
            True
            >>> range_1 == range_2
            True
            >>> range_1 == range_3
            False

            >>> range_2 == range_1
            True
            >>> range_2 == range_2
            True
            >>> range_2 == range_3
            False

            >>> range_3 == range_1
            False
            >>> range_3 == range_2
            False
            >>> range_3 == range_3
            True

        Returns true or false.
        """
        return super().__eq__(argument)

    def __format__(self, format_specification=''):
        """
        Formats pitch range.

        Set `format_specification` to `''` or `'storage'`.
        Interprets `''` equal to `'storage'`.

        Returns string.
        """
        if format_specification in ('', 'storage'):
            return StorageFormatManager(self).get_storage_format()
        return str(self)

    def __hash__(self):
        """
        Hashes pitch range.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __illustrate__(self):
        r"""
        Illustrates pitch range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange('[C3, C7]')
            >>> abjad.show(pitch_range) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = pitch_range.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Score])
                \new Score
                \with
                {
                    \override BarLine.stencil = ##f
                    \override Glissando.thickness = #2
                    \override SpanBar.stencil = ##f
                    \override TimeSignature.stencil = ##f
                }
                <<
                    \new PianoStaff
                    <<
                        \context Staff = "Treble Staff"
                        {
                            \clef "treble"
                            s1 * 1/4
                            s1 * 1/4
                        }
                        \context Staff = "Bass Staff"
                        {
                            \clef "bass"
                            c1 * 1/4
                            \glissando
                            \change Staff = "Treble Staff"
                            c''''1 * 1/4
                        }
                    >>
                >>

        Returns LilyPond file.
        """
        import abjad
        start_pitch_clef = abjad.Clef.from_selection(self.start_pitch)
        stop_pitch_clef = abjad.Clef.from_selection(self.stop_pitch)
        start_note = abjad.Note(self.start_pitch, 1)
        stop_note = abjad.Note(self.stop_pitch, 1)
        glissando = abjad.Glissando()
        if start_pitch_clef == stop_pitch_clef:
            if start_pitch_clef == abjad.Clef('bass'):
                bass_staff = abjad.Staff()
                abjad.attach(abjad.Clef('bass'), bass_staff)
                bass_staff.extend([start_note, stop_note])
                bass_leaves = abjad.select(bass_staff).leaves()
                abjad.attach(glissando, bass_leaves)
                score = abjad.Score([bass_staff])
            else:
                treble_staff = abjad.Staff()
                abjad.attach(abjad.Clef('treble'), treble_staff)
                treble_staff.extend([start_note, stop_note])
                treble_leaves = abjad.select(treble_staff).leaves()
                abjad.attach(glissando, treble_leaves)
                score = abjad.Score([treble_staff])
        else:
            result = abjad.Score.make_piano_score()
            score, treble_staff, bass_staff = result
            bass_staff.extend([start_note, stop_note])
            treble_staff.extend(abjad.Skip(1) * 2)
            bass_leaves = abjad.select(bass_staff).leaves()
            abjad.attach(glissando, bass_leaves)
            abjad.attach(abjad.StaffChange(treble_staff), bass_staff[1])
            abjad.attach(abjad.Clef('treble'), treble_staff[0])
            abjad.attach(abjad.Clef('bass'), bass_staff[0])
        for leaf in abjad.iterate(score).leaves():
            abjad.attach(abjad.Multiplier(1, 4), leaf)
        abjad.override(score).bar_line.stencil = False
        abjad.override(score).span_bar.stencil = False
        abjad.override(score).glissando.thickness = 2
        abjad.override(score).time_signature.stencil = False
        lilypond_file = abjad.LilyPondFile.new(score)
        return lilypond_file

    def __lt__(self, argument):
        """
        Is true when start pitch of this pitch-range is less than start
        pitch of `argument` pitch range.

        ..  container:: example

            >>> range_1 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_2 = abjad.PitchRange.from_pitches(-39, 0)
            >>> range_3 = abjad.PitchRange.from_pitches(-39, 48)

            >>> range_1 < range_1
            False
            >>> range_1 < range_2
            False
            >>> range_1 < range_3
            True

            >>> range_2 < range_1
            False
            >>> range_2 < range_2
            False
            >>> range_2 < range_3
            True

            >>> range_3 < range_1
            False
            >>> range_3 < range_2
            False
            >>> range_3 < range_3
            False

        Returns true or false.
        """
        try:
            argument = type(self)(argument)
        except (TypeError, ValueError):
            return False
        if self.start_pitch == argument.start_pitch:
            return self.stop_pitch < argument.stop_pitch
        return self.start_pitch < argument.start_pitch

    ### PRIVATE PROPERTIES ###

    @property
    def _close_bracket_string(self):
        if self.stop_pitch_is_included_in_range:
            return ']'
        else:
            return ')'

    @property
    def _open_bracket_string(self):
        if self.start_pitch_is_included_in_range:
            return '['
        else:
            return '('

    ### PRIVATE METHODS ###

    def _contains_pitch(self, pitch):
        import abjad
        if isinstance(pitch, numbers.Number):
            pitch = abjad.NamedPitch(pitch)
        elif isinstance(pitch, str):
            pitch = abjad.NamedPitch(pitch)
        if self.start_pitch is None and self.stop_pitch is None:
            return True
        elif self.start_pitch is None:
            if self.stop_pitch_is_included_in_range:
                return pitch <= self.stop_pitch
            else:
                return pitch < self.stop_pitch
        elif self.stop_pitch is None:
            if self.start_pitch_is_included_in_range:
                return self.start_pitch <= pitch
            else:
                return self.start_pitch < pitch
        else:
            if self.start_pitch_is_included_in_range:
                if self.stop_pitch_is_included_in_range:
                    return self.start_pitch <= pitch <= self.stop_pitch
                else:
                    return self.start_pitch <= pitch < self.stop_pitch
            else:
                if self.stop_pitch_is_included_in_range:
                    return self.start_pitch < pitch <= self.stop_pitch
                else:
                    return self.start_pitch < pitch < self.stop_pitch

    def _get_format_specification(self):
        return FormatSpecification(
            self,
            coerce_for_equality=True,
            repr_is_indented=False,
            storage_format_args_values=[self.range_string],
            storage_format_is_indented=False,
            )

    def _get_named_range_string(self):
        result = []
        result.append(self._open_bracket_string)
        if self.start_pitch:
            result.append(self.start_pitch.get_name(locale='us'))
        else:
            result.append('-inf')
        result.append(', ')
        if self.stop_pitch:
            result.append(self.stop_pitch.get_name(locale='us'))
        else:
            result.append('+inf')
        result.append(self._close_bracket_string)
        result = ''.join(result)
        return result

    def _get_numbered_range_string(self):
        result = []
        result.append(self._open_bracket_string)
        result.append(str(self.start_pitch.number))
        result.append(', ')
        result.append(str(self.stop_pitch.number))
        result.append(self._close_bracket_string)
        result = ''.join(result)
        return result

    def _list_numeric_octave_transpositions(self, pitch_number_list):
        result = []
        pitch_number_set = set(pitch_number_list)
        start_pitch_number = self.start_pitch.number
        stop_pitch_number = self.stop_pitch.number
        range_set = set(range(start_pitch_number, stop_pitch_number + 1))
        while pitch_number_set.issubset(range_set):
            next_pitch_number = list(pitch_number_set)
            next_pitch_number.sort()
            result.extend([next_pitch_number])
            pitch_number_set = set([_ + 12 for _ in pitch_number_set])
        pitch_number_set = set([_ - 12 for _ in pitch_number_list])
        while pitch_number_set.issubset(range_set):
            next_pitch_number = list(pitch_number_set)
            next_pitch_number.sort()
            result.extend([next_pitch_number])
            pitch_number_set = set([_ - 12 for _ in pitch_number_set])
        result.sort()
        return result

    def _parse_range_string(self, range_string):
        import abjad
        assert isinstance(range_string, str), repr(range_string)
        range_string = range_string.replace('-inf', '-1000')
        range_string = range_string.replace('+inf', '1000')
        match = constants._range_string_regex.match(range_string)
        if match is None:
            message = 'can not instantiate pitch range: {!r}'
            message = message.format(range_string)
            raise ValueError(message)
        group_dict = match.groupdict()
        start_punctuation = group_dict['open_bracket']
        start_pitch_string = group_dict['start_pitch']
        stop_pitch_string = group_dict['stop_pitch']
        stop_punctuation = group_dict['close_bracket']
        start_inclusivity_string = \
            constants._start_punctuation_to_inclusivity_string[start_punctuation]
        stop_inclusivity_string = \
            constants._stop_punctuation_to_inclusivity_string[stop_punctuation]
        if start_pitch_string == '-1000':
            start_pitch = None
        else:
            try:
                start_pitch = abjad.NamedPitch(start_pitch_string)
            except (TypeError, ValueError):
                start_pitch = abjad.NumberedPitch(int(start_pitch_string))
        if stop_pitch_string == '1000':
            stop_pitch = None
        else:
            try:
                stop_pitch = abjad.NamedPitch(stop_pitch_string)
            except (TypeError, ValueError):
                stop_pitch = abjad.NumberedPitch(int(stop_pitch_string))
        start_pair = (start_pitch, start_inclusivity_string)
        stop_pair = (stop_pitch, stop_inclusivity_string)
        return start_pair, stop_pair

    ### PUBLIC PROPERTIES ###

    @property
    def range_string(self):
        """
        Gets range string of pitch range.

        ..  container:: example

            >>> pitch_range = abjad.PitchRange('[C3, C7]')
            >>> pitch_range.range_string
            '[C3, C7]'

        Returns string.
        """
        return self._get_named_range_string()

    @property
    def start_pitch(self):
        """
        Start pitch of pitch range.

        >>> pitch_range = abjad.PitchRange('[C3, C7]')
        >>> pitch_range.start_pitch
        NamedPitch('c')

        Returns pitch.
        """
        if self._start is None:
            return None
        return self._start[0]

    @property
    def start_pitch_is_included_in_range(self):
        """
        Is true when start pitch is included in range.

        >>> pitch_range = abjad.PitchRange('[C3, C7]')
        >>> pitch_range.start_pitch_is_included_in_range
        True

        Returns true or false.
        """
        if self._start is None:
            return True
        return self._start[1] == 'inclusive'

    @property
    def stop_pitch(self):
        """
        Stop pitch of pitch range.

        >>> pitch_range = abjad.PitchRange('[C3, C7]')
        >>> pitch_range.stop_pitch
        NamedPitch("c''''")

        Returns pitch.
        """
        if self._stop is None:
            return None
        return self._stop[0]

    @property
    def stop_pitch_is_included_in_range(self):
        """
        Is true when stop pitch is included in range.

        >>> pitch_range = abjad.PitchRange('[C3, C7]')
        >>> pitch_range.stop_pitch_is_included_in_range
        True

        Returns true or false.
        """
        if self._stop is None:
            return True
        return self._stop[1] == 'inclusive'

    ### PUBLIC METHODS ###

    @staticmethod
    def from_pitches(
        start_pitch,
        stop_pitch,
        start_pitch_is_included_in_range=True,
        stop_pitch_is_included_in_range=True,
        ):
        """
        Initializes pitch range from numbers.

        ..  container:: example

            >>> abjad.PitchRange.from_pitches(-18, 19)
            PitchRange('[F#2, G5]')

        Returns pitch range.
        """
        import abjad
        if start_pitch is None:
            start_pitch_string = '-inf'
        else:
            start_pitch_string = str(abjad.NamedPitch(start_pitch))
        if stop_pitch is None:
            stop_pitch_string = '+inf'
        else:
            stop_pitch_string = str(abjad.NamedPitch(stop_pitch))
        start_containment = '['
        if not start_pitch_is_included_in_range:
            start_containment = '('
        stop_containment = ']'
        if not stop_pitch_is_included_in_range:
            stop_containment = ')'
        string = '{}{}, {}{}'
        string = string.format(
            start_containment,
            start_pitch_string,
            stop_pitch_string,
            stop_containment,
            )
        pitch_range = PitchRange(string)
        return pitch_range

    @classmethod
    def is_range_string(class_, argument):
        """Is true when `argument` is a pitch range string.

        ..  container:: example

            >>> abjad.PitchRange.is_range_string('[A0, C8]')
            True

            >>> abjad.PitchRange.is_range_string('[A#0, Cb~8]')
            True

            >>> abjad.PitchRange.is_range_string('[A#+0, cs'')')
            True

            >>> abjad.PitchRange.is_range_string('(b,,,, ctqs]')
            True

        ..  container:: example

            >>> abjad.PitchRange.is_range_string('text')
            False

        The regex that underlies this predicate matches against two
        comma-separated pitches enclosed in some combination of square
        brackets and round parentheses.

        Returns true or false.
        """
        if not isinstance(argument, str):
            return False
        return bool(constants._range_string_regex.match(argument))

    def list_octave_transpositions(self, pitch_carrier):
        """
        Lists octave transpositions of `pitch_carrier` in pitch range.

        ..  container:: example

            Lists octave transpositions of three-pitch chord:

            >>> chord = abjad.Chord("<c' d' e'>4")
            >>> pitch_range = abjad.PitchRange.from_pitches(0, 48)
            >>> result = pitch_range.list_octave_transpositions(chord)

            >>> for chord in result:
            ...     chord
            ...
            Chord("<c' d' e'>4")
            Chord("<c'' d'' e''>4")
            Chord("<c''' d''' e'''>4")
            Chord("<c'''' d'''' e''''>4")

        Returns a list of `pitch_carrier` objects.
        """
        import abjad
        if isinstance(pitch_carrier, collections.Iterable):
            if all(isinstance(x, (int, float)) for x in pitch_carrier):
                return self._list_numeric_octave_transpositions(pitch_carrier)
        prototype = (abjad.Chord, abjad.PitchSet)
        if not isinstance(pitch_carrier, prototype):
            message = 'must be chord or pitch-set: {!r}'
            message = message.format(pitch_carrier)
            raise TypeError(message)
        result = []
        interval = abjad.NumberedInterval(-12)
        while True:
            pitch_carrier_copy = copy.copy(pitch_carrier)
            candidate = interval.transpose(pitch_carrier_copy)
            if candidate in self:
                result.append(candidate)
                interval -= 12
            else:
                break
        result.reverse()
        interval = abjad.NumberedInterval(0)
        while True:
            pitch_carrier_copy = copy.copy(pitch_carrier)
            candidate = interval.transpose(pitch_carrier_copy)
            if candidate in self:
                result.append(candidate)
                interval += abjad.NumberedInterval(12)
            else:
                break
        return result

    def voice_pitch_class(self, pitch_class):
        """
        Voices `pitch_class`:

        ..  container:: example

            Voices C three times:

            >>> pitch_range = abjad.PitchRange('[C4, C6]')
            >>> pitch_range.voice_pitch_class('c')
            (NamedPitch("c'"), NamedPitch("c''"), NamedPitch("c'''"))

        ..  container:: example

            Voices B two times:

            >>> pitch_range = abjad.PitchRange('[C4, C6]')
            >>> pitch_range.voice_pitch_class('b')
            (NamedPitch("b'"), NamedPitch("b''"))

        ..  container:: example

            Returns empty because B can not voice:

            >>> pitch_range = abjad.PitchRange('[C4, A4)')
            >>> pitch_range.voice_pitch_class('b')
            ()

        Returns tuple of zero or more named pitches.
        """
        import abjad
        named_pitch_class = abjad.NamedPitchClass(pitch_class)
        pair = (named_pitch_class.name, self.start_pitch.octave.number)
        named_pitch = abjad.NamedPitch(pair)
        result = []
        while named_pitch <= self.stop_pitch:
            if named_pitch in self:
                result.append(named_pitch)
            named_pitch += 12
        return tuple(result)



import itertools


def yield_all_compositions_of_integer(n):
    """
    Yields all compositions of positive integer ``n``.

    ..  container:: example

        >>> for tuple_ in abjad.mathtools.yield_all_compositions_of_integer(5):
        ...     tuple_
        ...
        (5,)
        (4, 1)
        (3, 2)
        (3, 1, 1)
        (2, 3)
        (2, 2, 1)
        (2, 1, 2)
        (2, 1, 1, 1)
        (1, 4)
        (1, 3, 1)
        (1, 2, 2)
        (1, 2, 1, 1)
        (1, 1, 3)
        (1, 1, 2, 1)
        (1, 1, 1, 2)
        (1, 1, 1, 1, 1)

    Lists parts in descending lex order.

    Parts sum to ``n``.

    Finds small values of ``n`` easily.

    Takes around 4 seconds for ``n`` equal to 17.

    Returns integer tuple generator.
    """
    from abjad import mathtools
    compositions = []
    integer = 0
    string_length = n
    while integer < 2 ** (n - 1):
        binary_string = mathtools.integer_to_binary_string(integer)
        binary_string = binary_string.zfill(string_length)
        digits = [int(_) for _ in list(binary_string)]
        partition = []
        generator = itertools.groupby(digits, lambda _: _)
        for value, group in generator:
            partition.append(list(group))
        sublengths = [len(part) for part in partition]
        composition = tuple(sublengths)
        compositions.append(composition)
        integer += 1
    for composition in reversed(sorted(compositions)):
        yield composition



from .Segment import Segment


class PitchSegment(Segment):
    r"""
    Pitch segment.

    ..  container:: example

        Numbered pitch segment:

        >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

        >>> str(segment)
        '<-2, -1.5, 6, 7, -1.5, 7>'

        >>> abjad.show(segment) # doctest: +SKIP

        ..  docs::

            >>> lilypond_file = segment.__illustrate__()
            >>> abjad.f(lilypond_file[abjad.StaffGroup])
            \new PianoStaff
            <<
                \context Staff = "Treble Staff"
                {
                    \clef "treble"
                    r1 * 1/8
                    r1 * 1/8
                    fs'1 * 1/8
                    g'1 * 1/8
                    r1 * 1/8
                    g'1 * 1/8
                }
                \context Staff = "Bass Staff"
                {
                    \clef "bass"
                    bf1 * 1/8
                    bqf1 * 1/8
                    r1 * 1/8
                    r1 * 1/8
                    bqf1 * 1/8
                    r1 * 1/8
                }
            >>

    ..  container:: example

        Named pitch segment:

        >>> segment = abjad.PitchSegment("bf, aqs fs' g' bqf g'")

        >>> str(segment)
        "<bf, aqs fs' g' bqf g'>"

        >>> abjad.show(segment) # doctest: +SKIP

        ..  docs::

            >>> lilypond_file = segment.__illustrate__()
            >>> abjad.f(lilypond_file[abjad.StaffGroup])
            \new PianoStaff
            <<
                \context Staff = "Treble Staff"
                {
                    \clef "treble"
                    r1 * 1/8
                    r1 * 1/8
                    fs'1 * 1/8
                    g'1 * 1/8
                    r1 * 1/8
                    g'1 * 1/8
                }
                \context Staff = "Bass Staff"
                {
                    \clef "bass"
                    bf,1 * 1/8
                    aqs1 * 1/8
                    r1 * 1/8
                    r1 * 1/8
                    bqf1 * 1/8
                    r1 * 1/8
                }
            >>

    ..  container:: example

        Built-in max() works:

        >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])
        >>> max(segment)
        NumberedPitch(7)

        Built-in min() works:

        >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])
        >>> min(segment)
        NumberedPitch(-2)

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        if not items and not item_class:
            item_class = self._named_item_class
        Segment.__init__(
            self,
            items=items,
            item_class=item_class,
            )

    ### SPECIAL METHODS ###

    def __contains__(self, argument):
        """
        Is true when pitch segment contains `argument`.

        ..  container:: example

            Numbered pitch segment:

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.NamedPitch('fs') in segment
            False

            >>> 6 in segment
            True

            >>> abjad.NamedPitch('f') in segment
            False

            >>> 5 in segment
            False

        Returns true or false.
        """
        return super().__contains__(argument)

    def __illustrate__(self):
        r"""
        Illustrates pitch segment.

        ..  container:: example

            >>> segment = abjad.PitchSegment("bf, aqs fs' g' bqf g'")

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf,1 * 1/8
                        aqs1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

        Returns LilyPond file.
        """
        import abjad
        named_pitches = [abjad.NamedPitch(x) for x in self]
        maker = abjad.NoteMaker()
        notes = maker(named_pitches, [1])
        result = abjad.Score.make_piano_score(leaves=notes, sketch=True)
        score, treble_staff, bass_staff = result
        for leaf in abjad.iterate(score).leaves():
            abjad.attach(abjad.Multiplier(1, 8), leaf)
        abjad.override(score).rest.transparent = True
        lilypond_file = abjad.LilyPondFile.new(score)
        return lilypond_file

    def __repr__(self):
        """
        Gets interpreter representation of segment.

        Returns string.
        """
        import abjad
        if self.item_class is abjad.NamedPitch:
            contents = ' '.join([str(_) for _ in self])
            contents = '"' + contents + '"'
        else:
            contents = ', '.join([str(_) for _ in self])
            contents = '[' + contents + ']'
        return '{}({})'.format(type(self).__name__, contents)

    def __str__(self):
        """
        Gets pitch segment string.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> str(segment)
            '<-2, -1.5, 6, 7, -1.5, 7>'

        ..  container:: example

            >>> segment = abjad.PitchSegment("bf, aqs fs' g' bqf g'")

            >>> str(segment)
            "<bf, aqs fs' g' bqf g'>"

        Returns string.
        """
        import abjad
        items = [str(_) for _ in self]
        separator = ' '
        if self.item_class is abjad.NumberedPitch:
            separator = ', '
        return '<{}>'.format(separator.join(items))

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        import abjad
        return abjad.NamedPitch

    @property
    def _numbered_item_class(self):
        import abjad
        return abjad.NumberedPitch

    @property
    def _parent_item_class(self):
        import abjad
        return abjad.Pitch

    ### PRIVATE METHODS ###

    def _is_equivalent_under_transposition(self, argument):
        import abjad
        if not isinstance(argument, type(self)):
            return False
        if not len(self) == len(argument):
            return False
        difference = -(
            abjad.NamedPitch(argument[0], 4) -
            abjad.NamedPitch(self[0], 4)
            )
        new_pitches = (x + difference for x in self)
        new_pitches = abjad.new(self, items=new_pitches)
        return argument == new_pitches

    ### PUBLIC PROPERTIES ###

    @property
    def hertz(self):
        """
        Gets Hertz of pitches in segment.

        ..  container:: example

            >>> segment = abjad.PitchSegment('c e g b')
            >>> segment.hertz
            [130.81..., 164.81..., 195.99..., 246.94...]

        Returns list.
        """
        return [_.hertz for _ in self]

    @property
    def inflection_point_count(self):
        r"""
        Gets segment inflection point count.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment.inflection_point_count
            2

        Returns nonnegative integer.
        """
        return len(self.local_minima) + len(self.local_maxima)

    @property
    def local_maxima(self):
        r"""
        Gets segment local maxima.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment.local_maxima
            [NumberedPitch(7)]

        Returns list.
        """
        result = []
        if 3 <= len(self):
            for i in range(1, len(self) - 1):
                left, middle, right = self[i - 1], self[i], self[i + 1]
                if left < middle and right < middle:
                    result.append(middle)
        return result

    @property
    def local_minima(self):
        r"""
        Gets segment local minima.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment.local_minima
            [NumberedPitch(-1.5)]

        Returns list.
        """
        result = []
        if 3 <= len(self):
            for i in range(1, len(self) - 1):
                left, middle, right = self[i - 1], self[i], self[i + 1]
                if middle < left and middle < right:
                    result.append(middle)
        return result

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(
        class_,
        selection,
        item_class=None,
        ):
        r"""
        Makes pitch segment from `selection`.

        ..  container:: example

            >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = abjad.Staff("c4. r8 g2")
            >>> selection = abjad.select((staff_1, staff_2))
            >>> segment = abjad.PitchSegment.from_selection(selection)

            >>> str(segment)
            "<c' d' fs' a' b c g>"

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        c'1 * 1/8
                        d'1 * 1/8
                        fs'1 * 1/8
                        a'1 * 1/8
                        b1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        c1 * 1/8
                        g1 * 1/8
                    }
                >>

        Returns pitch segment.
        """
        import abjad
        if not isinstance(selection, abjad.Selection):
            selection = abjad.select(selection)
        named_pitches = []
        prototype = (abjad.Note, abjad.Chord)
        for component in abjad.iterate(selection).components(prototype):
            try:
                named_pitches.extend(component.written_pitches)
            except AttributeError:
                pass
            try:
                named_pitches.append(component.written_pitch)
            except AttributeError:
                pass
        return class_(
            items=named_pitches,
            item_class=item_class,
            )

    def has_duplicates(self):
        """
        Is true when segment has duplicates.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])
            >>> segment.has_duplicates()
            True

        ..  container:: example

            >>> segment = abjad.PitchSegment("c d e f g a b")
            >>> segment.has_duplicates()
            False

        Returns true or false.
        """
        import abjad
        return len(abjad.PitchSet(self)) < len(self)

    def invert(self, axis=None):
        r"""
        Inverts pitch segment about `axis`.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.invert(axis=0)

            >>> str(segment)
            '<2, 1.5, -6, -7, 1.5, -7>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        d'1 * 1/8
                        dqf'1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        dqf'1 * 1/8
                        r1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        r1 * 1/8
                        r1 * 1/8
                        fs1 * 1/8
                        f1 * 1/8
                        r1 * 1/8
                        f1 * 1/8
                    }
                >>

        Returns new pitch segment.
        """
        import abjad
        items = [_.invert(axis=axis) for _ in self]
        return abjad.new(self, items=items)

    def make_notes(self, n=None, written_duration=None):
        r"""
        Makes first `n` notes in pitch segment.

        ..  todo:: Move somewhere else.

        Sets `n` equal to `n` or length of segment.

        Sets `written_duration` equal to `written_duration` or ``1/8``:

        ..  container:: example

                >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> notes = segment.make_notes()
            >>> staff = abjad.Staff(notes)
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    bf8
                    bqf8
                    fs'8
                    g'8
                    bqf8
                    g'8
                }

        Allows nonassignable `written_duration`:

        ..  container:: example

            >>> notes = segment.make_notes(4, abjad.Duration(5, 16))
            >>> staff = abjad.Staff(notes)
            >>> time_signature = abjad.TimeSignature((5, 4))
            >>> abjad.attach(time_signature, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    \time 5/4
                    bf4
                    ~
                    bf16
                    bqf4
                    ~
                    bqf16
                    fs'4
                    ~
                    fs'16
                    g'4
                    ~
                    g'16
                }

        Returns list of notes.
        """
        import abjad
        n = n or len(self)
        written_duration = written_duration or abjad.Duration(1, 8)
        maker = abjad.NoteMaker()
        result = maker([0] * n, [written_duration])
        logical_ties = abjad.iterate(result).logical_ties()
        for i, logical_tie in enumerate(logical_ties):
            pitch = self[i % len(self)]
            for note in logical_tie:
                note.written_pitch = pitch
        return result

    def multiply(self, n=1):
        r"""
        Multiplies pitch segment by index `n`.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.multiply(n=3)

            >>> str(segment)
            '<-6, -4.5, 18, 21, -4.5, 21>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs''1 * 1/8
                        a''1 * 1/8
                        r1 * 1/8
                        a''1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        fs1 * 1/8
                        gqs1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        gqs1 * 1/8
                        r1 * 1/8
                    }
                >>

        Returns new pitch segment.
        """
        import abjad
        items = [_.multiply(n=n) for _ in self]
        return abjad.new(self, items=items)

    def retrograde(self):
        r"""
        Retrograde of pitch segment.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.retrograde()

            >>> str(segment)
            '<7, -1.5, 7, 6, -1.5, -2>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                        fs'1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        bf1 * 1/8
                    }
                >>

        Returns new pitch segment.
        """
        import abjad
        return abjad.new(self, items=reversed(self))

    def rotate(self, n=0, stravinsky=False):
        r"""
        Rotates pitch segment by index `n`.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.rotate(n=1)

            >>> str(segment)
            '<7, -2, -1.5, 6, 7, -1.5>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        g'1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        r1 * 1/8
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                    }
                >>

        Returns new pitch segment.
        """
        import abjad
        rotated_pitches = abjad.sequence(self._collection).rotate(n=n)
        new_segment = abjad.new(self, items=rotated_pitches)
        if stravinsky:
            if self[0] != new_segment[0]:
                interval = new_segment[0] - self[0]
                new_segment = new_segment.transpose(interval)
        return new_segment

    def to_pitch_classes(self):
        r"""
        Changes to pitch-classes.

        ..  container:: example

            To numbered pitch-class segment:

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.to_pitch_classes()

            >>> str(segment)
            'PC<10, 10.5, 6, 7, 10.5, 7>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    bf'8
                    bqf'8
                    fs'8
                    g'8
                    bqf'8
                    g'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        ..  container:: example

            To named pitch-class segment:

            >>> segment = abjad.PitchSegment("bf, aqs fs' g' bqf g'")

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf,1 * 1/8
                        aqs1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.to_pitch_classes()

            >>> str(segment)
            'PC<bf aqs fs g bqf g>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    bf'8
                    aqs'8
                    fs'8
                    g'8
                    bqf'8
                    g'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        Returns new segment.
        """
        import abjad
        class_ = abjad.Pitch
        item_class = class_._to_pitch_class_item_class(self.item_class)
        return abjad.PitchClassSegment(items=self.items, item_class=item_class)

    def to_pitches(self):
        r"""
        Changes to pitches.

        ..  container:: example

            To numbered pitch segment:

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.to_pitches()

            >>> str(segment)
            '<-2, -1.5, 6, 7, -1.5, 7>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

        ..  container:: example

            To named pitch segment:

            >>> segment = abjad.PitchSegment("bf, aqs fs' g' bqf g'")

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf,1 * 1/8
                        aqs1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.to_pitches()

            >>> str(segment)
            "<bf, aqs fs' g' bqf g'>"

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf,1 * 1/8
                        aqs1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

        Returns new segment.
        """
        import abjad
        return abjad.new(self)

    def transpose(self, n=0):
        r"""
        Transposes pitch segment by index `n`.

        ..  container:: example

            >>> segment = abjad.PitchSegment([-2, -1.5, 6, 7, -1.5, 7])

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        r1 * 1/8
                        r1 * 1/8
                        fs'1 * 1/8
                        g'1 * 1/8
                        r1 * 1/8
                        g'1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        bf1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        bqf1 * 1/8
                        r1 * 1/8
                    }
                >>

            >>> segment = segment.transpose(n=11)

            >>> str(segment)
            '<9, 9.5, 17, 18, 9.5, 18>'

            >>> abjad.show(segment) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.StaffGroup])
                \new PianoStaff
                <<
                    \context Staff = "Treble Staff"
                    {
                        \clef "treble"
                        a'1 * 1/8
                        aqs'1 * 1/8
                        f''1 * 1/8
                        fs''1 * 1/8
                        aqs'1 * 1/8
                        fs''1 * 1/8
                    }
                    \context Staff = "Bass Staff"
                    {
                        \clef "bass"
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                        r1 * 1/8
                    }
                >>

        Returns new pitch segment.
        """
        import abjad
        items = [_.transpose(n=n) for _ in self]
        return abjad.new(self, items=items)



"""
Utilities for typehinting.
"""

import typing
from abjad.utilities.Expression import Expression


IntegerPair = typing.Tuple[int, int]

Number = typing.Union[int, float]

NumberPair = typing.Tuple[Number, Number]

Selector = typing.Union[str, Expression]



import abc
from abjad.system.AbjadObject import AbjadObject


class TypedCollection(AbjadObject):
    """
    Abstract typed collection.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_collection',
        '_item_class',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(
        self,
        items=None,
        item_class=None,
        ):
        assert isinstance(item_class, (type(None), type))
        self._item_class = item_class

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        """
        Is true when typed collection contains ``item``.

        Returns true or false.
        """
        try:
            item = self._item_coercer(item)
        except ValueError:
            return False
        return self._collection.__contains__(item)

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a typed collection with items that
        compare equal to those of this typed collection.

        Returns true or false.
        """
        if issubclass(type(argument), type(self)):
            return self._collection == argument._collection
        elif isinstance(argument, type(self._collection)):
            return self._collection == argument
        return False

    def __hash__(self):
        """
        Hashes typed collection.

        Redefined in tandem with __eq__.
        """
        return object.__hash__(self)

    def __iter__(self):
        """
        Iterates typed collection.

        Returns generator.
        """
        return self._collection.__iter__()

    def __len__(self):
        """
        Gets length of typed collection.

        Returns nonnegative integer.
        """
        return len(self._collection)

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        def coerce_(x):
            if isinstance(x, self._item_class):
                return x
            return self._item_class(x)
        if self._item_class is None:
            return lambda x: x
        return coerce_

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        import abjad
        agent = abjad.StorageFormatManager(self)
        names = list(agent.signature_keyword_names)
        if 'items' in names:
            names.remove('items')
        return abjad.FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self._collection],
            storage_format_kwargs_names=names,
            )

    def _on_insertion(self, item):
        """
        Override to operate on item after insertion into collection.
        """
        pass

    def _on_removal(self, item):
        """
        Override to operate on item after removal from collection.
        """
        pass

    ### PUBLIC PROPERTIES ###

    @property
    def item_class(self):
        """
        Gets item class of collection.

        Collection coerces items according to ``item_class``.

        Returns class.
        """
        return self._item_class

    @property
    def items(self):
        """
        Gets items in collection.

        Returns list.
        """
        return [_ for _ in self]



import math


def binomial_coefficient(n, k):
    """
    Gets binomial coefficient of ``n`` choose ``k``.

    ..  container:: example

        >>> for k in range(8):
        ...     print(k, '\t', abjad.mathtools.binomial_coefficient(8, k))
        ...
        0  1
        1  8
        2  28
        3  56
        4  70
        5  56
        6  28
        7  8

    Returns positive integer.
    """
    return math.factorial(n) // (math.factorial(n - k) * math.factorial(k))
